#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>

#define BIT_TMS	0
#define BIT_TCK 1
#define BIT_TDI	2

#define BIT_TDO 5

#define BIT(n)  (1UL << (n))

typedef unsigned char  U8;
typedef unsigned short U16;
typedef unsigned long  U32;
typedef signed char    S8;
typedef signed short   S16;
typedef signed long    S32;

#if 0		 /* old Linux code */
#include <unistd.h>
#include <sys/io.h>
#include <sys/perm.h>

#define BASE      0x378                /* where LPT1: is on most h/w */
#define LP_DATA   (BASE+0)             /* write to this to change D0-D7 */
#define LP_STATUS (BASE+1)             /* read from this to get status */
#define LP_CTL    (BASE+2)             /* write to change the ctl lines */


static void do_init(void)
{
  if (ioperm(BASE,3,1)) {
    fprintf(stderr, "Couldn't get the port at %x\n", BASE);
    exit(1);
  }
}

static void do_out(unsigned long outval)
{
  //printf("sending %02lx\n", outval);
  outb(outval, LP_DATA);
}

static unsigned long do_in(void)
{
  unsigned int i = inb(LP_STATUS);;

  return (i);
}

static unsigned long do_jtag_in(void)
{
  return (1 & (do_in() >> BIT_TDO));
}
#else
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include </sys/dev/ppbus/ppi.h>
#include </sys/dev/ppbus/ppbconf.h>

static int ppi_fd;
static FILE *outlog;

static void do_init(void)
{
  char port[] = "/dev/ppi0";

  ppi_fd = open(port, O_RDWR);
  if( ppi_fd < 0 ) {
    perror(port);
    exit(1);
  }
  outlog = fopen("out", "w");
}

static void do_out(unsigned long outval)
{
  static unsigned long prev_outval;
  int val = outval, n;

  if ((~prev_outval & outval) & BIT(BIT_TCK)) {
    if (outlog) {
      fprintf(outlog, "TMS %d, TDI %d\n",
              1 & (outval >> BIT_TMS),
              1 & (outval >> BIT_TDI));
    }
  }
  n = ioctl(ppi_fd, PPISDATA, &val);
  if( n < 0 ) {
    perror("ioctl PPISDATA");
    exit(1);
  }
  prev_outval = outval;
}

static unsigned long do_in(void)
{
  int val, n;

  n = ioctl(ppi_fd, PPIGSTATUS, &val);
  //printf("Read %02x\n", val);
  if( n < 0 ) {
    perror("ioctl PPIGSTATUS");
    exit(1);
  }
  return val;
}

static unsigned long do_jtag_in(void)
{
  return (1 & (do_in() >> BIT_TDO));
}


#endif

/************************************************************************/

struct state {
  char *name;
  struct state *zero, *one;
};

#define TEST_LOGIC_RESET	0
#define RUN_TEST_IDLE		1
#define SELECT_DR_SCAN		2
#define CAPTURE_DR		3
#define SHIFT_DR		4
#define EXIT1_DR		5
#define PAUSE_DR		6
#define EXIT2_DR		7
#define UPDATE_DR		8
#define SELECT_IR_SCAN		9
#define CAPTURE_IR		10
#define SHIFT_IR		11
#define EXIT1_IR		12
#define PAUSE_IR		13
#define EXIT2_IR		14
#define UPDATE_IR		15	

static struct state machine[] = {
  { "Test-Logic-Reset",    &machine[1],  &machine[0]  }, /* 0 */
  { "Run-Test/Idle",       &machine[1],  &machine[2]  }, /* 1 */
  { "Select-DR-Scan",      &machine[3],  &machine[9]  }, /* 2 */
  { "Capture-DR",          &machine[4],  &machine[5]  }, /* 3 */
  { "Shift-DR",            &machine[4],  &machine[5]  }, /* 4 */
  { "Exit1-DR",		   &machine[6],  &machine[8]  }, /* 5 */
  { "Pause-DR",            &machine[6],  &machine[7]  }, /* 6 */
  { "Exit2-DR",            &machine[4],  &machine[8]  }, /* 7 */
  { "Update-DR",           &machine[1],  &machine[2]  }, /* 8 */
  { "Select-IR-Scan",      &machine[10], &machine[0]  }, /* 9 */
  { "Capture-IR",          &machine[11], &machine[12] }, /* 10 */
  { "Shift-IR",            &machine[11], &machine[12] }, /* 11 */
  { "Exit1-IR",            &machine[13], &machine[15] }, /* 12 */
  { "Pause-IR",            &machine[13], &machine[14] }, /* 13 */
  { "Exit2-IR",            &machine[11], &machine[15] }, /* 14 */
  { "Update-IR",           &machine[1],  &machine[2]  } /* 15 */
  };

static struct state *current_state = &machine[0];

#define ASSERT_STATE(S) assert((current_state - machine) == (S))

#define STANDARD_DELAY 0

static void go_state(int tms)
{
  unsigned long mask = (tms << BIT_TMS);

  if (tms == 0)
    current_state = current_state->zero;
  else
    current_state = current_state->one;

  do_out(mask);
  do_out(mask | BIT(BIT_TCK));
}

static int do_bit(int tms, int tdi)
{
  unsigned long mask = (tdi << BIT_TDI) | (tms << BIT_TMS);
  int sample;

  if (tms == 0)
    current_state = current_state->zero;
  else
    current_state = current_state->one;

  do_out(mask);
  sample = (int)(do_jtag_in());
  do_out(mask | BIT(BIT_TCK));

  return sample;
}

static unsigned long long do_nbit(int n, unsigned long long tdi)
{
  unsigned long long mask;
  unsigned long long tdo;
  int i;

  tdo = 0;
  for (i = 0; i < n; i++) {
    int is_lastbit = (i == (n - 1));
    mask = 1ULL << i;
    
    if (do_bit(is_lastbit, (tdi & mask) ? 1 : 0))
      tdo |= mask;
  }

  return tdo;
}

static int do_xbit(int tms, int tdi)
{
  unsigned long mask = (tdi << BIT_TDI) | (tms << BIT_TMS);
  int sample0, sample1, sample2;

  if (tms == 0)
    current_state = current_state->zero;
  else
    current_state = current_state->one;

  //printf("tdi=%d, tms=%d\n", tdi, tms);
  sample0 = (int)(do_jtag_in());
  do_out(mask);
  sample1 = (int)(do_jtag_in());
  do_out(mask | BIT(BIT_TCK));
  sample2 = (int)(do_jtag_in());

  //printf("state is now %s, TDI=%d, TDO=%d\n", current_state->name, tdi, sample);
  //printf("%d %d %d\n", sample0, sample1, sample2);

  return sample2;
}

static unsigned long readN(int n)
{
  unsigned long long tdo;
  int i;

  tdo = 0;
  for (i = 0; i < (n + 1); i++) {
    tdo <<= 1;

    //printf("%2d: ", 32 - i);
    if (do_xbit(0, 0))
      tdo |= 1;
  }

  return tdo;
}

static void send_str(char *s, int end)
{
  for (; *s; s++) {
    if ((*s == '0') || (*s == '1')) {
      int is_lastbit = (end && (s[1] == '\0'));
    
      do_bit(is_lastbit, (*s == '1'));
    }
  }
}

static unsigned long long do_nbit_cycle(int n, unsigned long long tdi)
{
  unsigned long long mask;
  unsigned long long tdo;
  int i;

  tdo = 0;
  for (i = 0; i < n; i++) {
    mask = 1ULL << i;
    
    if (do_bit(0, (tdi & mask) ? 1 : 0))
      tdo |= mask;
  }

  return tdo;
}

/************************************************************************/

static int show_status(void)
{
  int status;

  go_state(1);
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(0);

  go_state(1);
  go_state(1);
  go_state(0);
  go_state(0);
  ASSERT_STATE(SHIFT_IR);
  do_nbit(5, 0x05);		/* CFG_IN */
  go_state(1);
  go_state(1);
  go_state(0);
  go_state(0);
  ASSERT_STATE(SHIFT_DR);
  send_str("0010 1000 0000 0000 1110 0000 0000 0001", 0);
  send_str("0000 0000 0000 0000 0000 0000 0000 0000", 1);
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(0);
  go_state(0);
  ASSERT_STATE(SHIFT_IR);
  do_nbit(5, 0x04);		/* CFG_OUT */
  go_state(1);
  go_state(1);
  go_state(0);
  go_state(0);
  ASSERT_STATE(SHIFT_DR);
  {
    status = readN(32);

    printf("status=%08x\n", status);
#if 0      
    for (i = 0; i < 32; i++)
      printf("%d %s", (status >> (31 - i)) & 1, ((i & 3) == 3) ? " " : "");
#endif

    printf("\n");
    printf("%10s: %d\n", "DONE", (status >> 14) & 1);
    printf("%10s: %d\n", "INIT", (status >> 13) & 1);
    printf("%10s: %d\n", "MODE", (status >> 10) & 7);
    printf("%10s: %d\n", "GHIGH_B", (status >> 9) & 1);
    printf("%10s: %d\n", "GSR_B", (status >> 8) & 1);
    printf("%10s: %d\n", "GWE_B", (status >> 7) & 1);
    printf("%10s: %d\n", "GTS_CFG", (status >> 6) & 1);
    printf("%10s: %d\n", "IN_ERROR", (status >> 5) & 1);
    printf("%10s: %d%d%d%d\n", "LOCK", 
	   (status >> 4) & 1,
	   (status >> 3) & 1,
	   (status >> 2) & 1,
	   (status >> 1) & 1);
    printf("%10s: %d\n", "CRC_ERROR", (status >> 0) & 1);
  }

  return ((status >> 14) != 0) && ((status & 1) == 0);
}

static void do_cfg_in(void)
{
  go_state(1);		/* 1 */
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(1);

  go_state(0);		/* 2 */

  go_state(1);		/* 3 */
  go_state(1);

  go_state(0);		/* 4 */
  go_state(0);

  ASSERT_STATE(SHIFT_IR);	/* 5 */
  do_nbit(5, 0x05);		/* CFG_IN */

  go_state(1);		/* 7 */
  go_state(1);
  go_state(0);		/* 8 */
  go_state(0);
  ASSERT_STATE(SHIFT_DR);
}

static void do_shutdown()
{
  int i;

  do_cfg_in();

#if 1
  send_str("0011 0000 0000 0001 0010 0000 0000 0001", 0);
  send_str("0000 0000 1010 0000 1011 1111 0010 1101", 0);
  send_str("0011 0000 0000 0000 1000 0000 0000 0001", 0);
  send_str("0000 0000 0000 0000 0000 0000 0000 0101", 0);
  send_str("0011 0000 0000 0000 1000 0000 0000 0001", 0);
  send_str("0000 0000 0000 0000 0000 0000 0000 0111", 0);
  send_str("0000 0000 0000 0000 0000 0000 0000 0000", 0);
  send_str("0000 0000 0000 0000 0000 0000 0000 0000", 1);

  go_state(1);
  go_state(1);
  go_state(1);
  go_state(0);
  go_state(0);
  ASSERT_STATE(SHIFT_IR);
  do_nbit(5, 0x0c);		/* JSTART */

  go_state(1);			/* 14 */
  go_state(1);
  go_state(0);
  go_state(0);
  ASSERT_STATE(SHIFT_DR);

  for (i = 0; i < 14; i++)
    go_state(0);

#if 0
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(0);
  go_state(0);
  ASSERT_STATE(SHIFT_IR);
  do_nbit(5, 0x05);		/* CFG_IN */
  go_state(1);			/* 7 */
  go_state(1);
  go_state(0);			/* 8 */
  go_state(0);
  ASSERT_STATE(SHIFT_DR);

  send_str("0011 0000 0000 0000 1000 0000 0000 0001", 0);
  send_str("0000 0000 0000 0000 0000 0000 0000 1000", 0);
  send_str("0011 0000 0000 0001 0010 0000 0000 0001", 0);
  send_str("0000 0000 1010 0000 0011 1111 0010 1101", 0);
  send_str("0000 0000 0000 0000 0000 0000 0000 0000", 1);
#endif
#endif

  go_state(1);		/* 1 */
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(1);
}

static void load_rbt(char *rbtfile)
{
  int i;
  FILE *rbt;
  char line[80];
  int nbits = 947;

  do_shutdown();
  //show_status();

  {
    do_cfg_in();

    rbt = fopen(rbtfile, "r");
    if (rbt == NULL) {
      perror(rbtfile);
      exit(1);
    }
    fgets(line, sizeof(line), rbt);
    fgets(line, sizeof(line), rbt);
    fgets(line, sizeof(line), rbt);
    fgets(line, sizeof(line), rbt);
    fgets(line, sizeof(line), rbt);
    fgets(line, sizeof(line), rbt);
    fgets(line, sizeof(line), rbt);
    sscanf(line, "Bits:        	%d\n", &nbits);
    printf("%d bits\n", nbits);

    fprintf(outlog, "====\n");
    for (; nbits > 0; nbits -= 32) {
      int islast;
      fgets(line, sizeof(line), rbt);
      *strchr(line, '\r') = '\0';
      islast = (nbits == 32);
      islast = 0;
#if 0
      printf("%s %04lx %04lx <%d>\n", 
	     line, 
	     strtoul(line, NULL, 2) >> 16,
	     strtoul(line, NULL, 2) & 0xffff,
	     islast);
#endif
      send_str(line, islast);
    }
    fprintf(outlog, "====\n");
    send_str("0", 1);

#if 1
    go_state(1);
    go_state(1);
    go_state(1);
    go_state(0);
    go_state(0);
    ASSERT_STATE(SHIFT_IR);
    do_nbit(5, 0x0c);		/* JSTART */

    go_state(1);		/* 14 */
    go_state(1);
    go_state(0);
    go_state(0);
    ASSERT_STATE(SHIFT_DR);

    for (i = 0; i < 14; i++)
      go_state(0);
    
    go_state(1);		/* 14 */
    go_state(1);
    ASSERT_STATE(UPDATE_DR);
    go_state(0);
    ASSERT_STATE(RUN_TEST_IDLE);
#endif
  }

  //if (!show_status())
  //  exit(1);
}

int main(int argc, char *argv[])
{
  if (argc == 1) {
    do_init();
    show_status();
    exit(0);
  }
  if (argc != 2) {
    fprintf(stderr, "usage: %s <rbt-file>\n", argv[0]);
    exit(1);
  }

  do_init();

  go_state(1);
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(1);
  go_state(0);
  
  ASSERT_STATE(RUN_TEST_IDLE);
  
  load_rbt(argv[1]);

  exit(0);
}
